<?php

/**
 * Contains MultiStepFormBase.
 */

namespace Drupal\multistep_form_framework\Form;

use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\HtmlCommand;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Theme\ThemeManagerInterface;

abstract class MultiStepFormBase extends FormBase implements MultiStepFormInterface {
	/**
	 * FormStep Class.
	 *
	 * @var string
	 */
	protected $formStep;
	
	/**
	 * Aggregate of current form values.
	 *
	 * @var array
	 */
	protected $formValues;

	/**
	 * The module handler.
	 *
	 * @var \Drupal\Core\Extension\ModuleHandlerInterface
	 */
	protected $moduleHandler;
	
	/**
     * The theme manager.
     *
     * @var \Drupal\Core\Theme\ThemeManagerInterface
     */
	protected $themeManager;
  
	/**
	 * Constructor
	 *
	 * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
	 *   The module handler.
     * @param \Drupal\Core\Theme\ThemeManagerInterface $theme_manager
     *   The theme manager.	 
	 *
	 * @return void
	 */
	public function __construct(ModuleHandlerInterface $module_handler, ThemeManagerInterface $theme_manager) {
		$this->moduleHandler = $module_handler;
		$this->themeManager = $theme_manager;

		$this->formStep = $this->getInitialFormStep();
		$this->formValues = [];
	}
	
	/**
	 * {@inheritdoc}
	 */
	public static function create(ContainerInterface $container) {
		return new static($container->get('module_handler'), $container->get('theme.manager'));
	}

	/**
	 * Gets the current form step class.
	 *
	 * @return string 
	 *   Current FormStep class.
	 */
	public function getFormStep() {
		return $this->formStep;
	}
	
	/**
	 * Sets the current form step class.
	 *
	 * @param string $formStep
	 *   FormStep class namespace path.
	 * 
	 * @return void
	 */
	public function setFormStep($formStep) {
		$this->formStep = $formStep;
	}
	
	/**
	 * {@inheritdoc}
	 */
	public function buildForm(array $form, FormStateInterface $form_state) {
		// Build form step.
		$formStep = [];
		$formStep = $this->buildCurrentFormStep($formStep, $form_state);

		// Create wrapper if this is an ajax form.
		if ($this->isAjaxForm()) {
			$form['wrapper_messages'] = [
				'#type' => 'container',
				'#attributes' => [
					'id' => $this->getAjaxWrapperId() . '_messages',
				],
			];
			
			$form['wrapper_messages'] += $this->buildMessages();
			
			$form['wrapper'] = [
			  '#type' => 'container',
			  '#attributes' => [
				'id' => $this->getAjaxWrapperId(),
			  ],
			];
		}
		
		// Add form step elements to the form array.
		if ($this->isAjaxForm()) {
			$form['wrapper'] += $formStep;
		} else {
			$form += $formStep;
		}
		
		return $form;
	}
	
	/**
	 * Return status message form element.
	 *
	 * @return array
	 */
	protected function buildMessages() {
		if ($messages = $this->messenger()->all()) {
			return [
				'#theme' => 'status_messages',
				'#message_list' => $messages,
				'#status_headings' => [
					'status' => $this->t('Status message'),
					'error' => $this->t('Error message'),
					'warning' => $this->t('Warning message')
				]
			];
		} else {
			return [];
		}
	}

	/**
	 * Ajax Callback Function
	 * 
     * @param array $form
     *   An associative array containing the structure of the form.
     * @param \Drupal\Core\Form\FormStateInterface $form_state
     *   The current state of the form.
	 *
	 * @return \Drupal\Core\Ajax\AjaxResponse
	 *   JSON response object for AJAX request.
	 */
	public function ajaxCallback(array &$form, FormStateInterface $form_state) {
		$response = new AjaxResponse();
		$response->addCommand(new HtmlCommand('#' . $this->getAjaxWrapperId(), $form['wrapper']));

		// Add status message element.
		$response->addCommand(new HtmlCommand('#' . $this->getAjaxWrapperId() . '_messages', $this->buildMessages()));

		return $response;
	}
	
	/**
	 * @{inheritdoc}
	 */
	public function getAjaxCallback() {
		return [$this, 'ajaxCallback'];
	}
	
	/**
	 * Returns and instance of the current FormStep class.
	 *
	 * @return /Drupal/multistep_form_framework/FormStep/FormStepInterface
	 *   Instance of current FormStep.
	 */
	protected function getCurrentFormObject() {
		$class = $this->getFormStep();
		$instance = ($class::create(\Drupal::getContainer()));
		$instance->setMultiStepForm($this);
		return $instance;
	}
	
	/**
	 * Instantiates FormStep and calls it's buildFormStep.
	 *
     * @param array $form
     *   An associative array containing the structure of the form.
     * @param \Drupal\Core\Form\FormStateInterface $form_state
     *   The current state of the form.
	 *
	 * @return array
	 *   The form structure.
	 */
	protected function buildCurrentFormStep(array &$form, FormStateInterface $form_state) {
		// Derived from \Drupal\Core\Form\FormBuilder::prepareForm
		$form_instance = $this->getCurrentFormObject();
		$form_id = $form_instance->getFormId();
		$hooks[] = 'form_' . $form_id;
		$form = $form_instance->buildForm($form, $form_state);
		
		// Invoke hook_form_alter(), hook_form_BASE_FORM_ID_alter(), and
		// hook_form_FORM_ID_alter() implementations.
		$this->moduleHandler
			->alter($hooks, $form, $form_state, $form_id);
		$this->themeManager
			->alter($hooks, $form, $form_state, $form_id);
		
		return $form;
	}
	
	/**
	 * Instantiates FormState class and calls it's validateFormStep.
	 *
     * @param array $form
     *   An associative array containing the structure of the form.
     * @param \Drupal\Core\Form\FormStateInterface $form_state
     *   The current state of the form.
	 */	
	protected function validateCurrentFormStep(array &$form, FormStateInterface $form_state) {
		$this->getCurrentFormObject()->validateForm($form, $form_state);
	}
	
    /**
	 * Instantiates FormState class and calls it's submitFormStep.
	 *
     * @param array $form
     *   An associative array containing the structure of the form.
     * @param \Drupal\Core\Form\FormStateInterface $form_state
     *   The current state of the form.
	 */	
	protected function submitCurrentFormStep(array &$form, FormStateInterface $form_state) {
		$this->getCurrentFormObject()->submitForm($form, $form_state);
	}
	
	/**
	 * {@inheritdoc}
	 */
	public function validateForm(array &$form, FormStateInterface $form_state) {
		$this->validateCurrentFormStep($form, $form_state);
		
		// Suppress required field validation errors if action is previous.

		$triggerElement = $form_state->getTriggeringElement();
		if (isset($triggerElement['#form_action']) && $triggerElement['#form_action'] == 'previous') {
			$validationErrors = $form_state->getErrors();
			$form_state->clearErrors();
			
			foreach ($validationErrors as $index => $validationError) {
				if (preg_match('/\w+ field is required./', $validationError->__toString()) === 0)
					$form_state->setError($index, $validationError);
			}
		}
	}

	/**
	 * {@inheritdoc}
	 */
	public function submitForm(array &$form, FormStateInterface $form_state) {
		$this->submitCurrentFormStep($form, $form_state);
		
		foreach ($form_state->getValues() as $index => $value) {
			$this->formValues[$index] = $value;
		}
		
		$form_state->setValues($this->formValues);
		
		$triggerElement = $form_state->getTriggeringElement();
		if (isset($triggerElement['#form_step'])) $this->setFormStep($triggerElement['#form_step']);
		
		if (isset($triggerElement['#form_complete']) && $triggerElement['#form_complete']) {
			$this->submitAllForm($form, $form_state);
		}
		
		$form_state->setRebuild(TRUE);
	}
	
	/**
	 * Submit callback for completed form.
	 *
     * @param array $form
     *   An associative array containing the structure of the form.
     * @param \Drupal\Core\Form\FormStateInterface $form_state
     *   The current state of the form.
	 */
	public function submitAllForm(array &$form, FormStateInterface $form_state) {}
}
